import pickle

from io import BytesIO
from io import StringIO
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from actionpack.action import Result
from actionpack.actions import Remove
from actionpack.utils import pickleable
from tests.actionpack.actions import uncloseable


class RemoveTest(TestCase):

    contents = 'some file contents.'
    separator = '\n'
    other_contents = f'{contents}{separator}'
    head = 'head'
    tail = 'tail'
    head_and_tail = f'{head}{separator}{tail}'

    def test_instantiate_given_invalid_output_type(self):
        result = Remove('some-file.name', output_type=100).perform()
        self.assertIsInstance(result.value, TypeError)

    @patch('builtins.open')
    def test_can_handle_empty_files(self, mock_file):
        with uncloseable(StringIO()) as buffer:
            mock_file.return_value = buffer

            action = Remove(__file__)
            result = action.perform()

            self.assertIsInstance(result, Result)
            self.assertTrue(result.successful)
            self.assertEqual(result.value, '')

    @patch('builtins.open')
    def test_can_Remove_string(self, mock_file):
        num_repeats = 5
        with uncloseable(StringIO((self.other_contents * num_repeats).rstrip())) as buffer:
            mock_file.return_value = buffer

            action = Remove(__file__)
            result = action.perform()

            self.assertIsInstance(result, Result)
            self.assertTrue(result.successful)
            self.assertEqual(result.value, self.other_contents)

        remaining_contents = buffer.read().split(self.separator)
        self.assertEqual(len(remaining_contents), num_repeats - 1)
        self.assertTrue(remaining == self.contents for remaining in remaining_contents)

    @patch('builtins.open')
    def test_can_Remove_string_from_file_head(self, mock_file):
        with uncloseable(StringIO(self.head_and_tail)) as buffer:
            mock_file.return_value = buffer

            action = Remove(__file__)
            result = action.perform()

            self.assertIsInstance(result, Result)
            self.assertTrue(result.successful)
            self.assertEqual(result.value, self.head + self.separator)

        remaining_contents = buffer.read().rstrip()
        self.assertEqual(remaining_contents, self.tail)

    @patch('builtins.open')
    def test_can_Remove_string_from_file_tail(self, mock_file):
        with uncloseable(StringIO(self.head_and_tail)) as buffer:
            mock_file.return_value = buffer

            action = Remove(__file__, tail=True)
            result = action.perform()

            self.assertIsInstance(result, Result)
            self.assertTrue(result.successful)
            self.assertEqual(result.value, self.tail)

        remaining_contents = buffer.read().rstrip()
        self.assertEqual(remaining_contents, self.head)

    @patch('builtins.open')
    def test_can_Remove_bytes(self, mock_file):
        num_repeats = 5
        with uncloseable(BytesIO((self.other_contents * num_repeats).rstrip().encode())) as buffer:
            mock_file.return_value = buffer

            action = Remove(__file__)
            result = action.perform()

            self.assertIsInstance(result, Result)
            self.assertTrue(result.successful)
            self.assertEqual(result.value, self.other_contents.encode())

        remaining_contents = buffer.read().split(self.separator.encode())
        self.assertEqual(len(remaining_contents), num_repeats - 1)
        self.assertTrue(remaining == self.contents.encode() for remaining in remaining_contents)

    @patch('builtins.open')
    def test_can_Remove_bytes_even_if_failure(self, mock_file):
        contents = self.contents.encode()
        mock_file.return_value = contents

        invalid_file_result = Remove('some/invalid/filepath').perform()
        self.assertIsInstance(invalid_file_result, Result)
        self.assertIsInstance(invalid_file_result.value, FileNotFoundError)

        directory_result = Remove(Path(__file__).parent).perform()
        self.assertIsInstance(directory_result, Result)
        self.assertIsInstance(directory_result.value, IsADirectoryError)

    def test_can_pickle(self):
        action = Remove(__file__)
        pickled = pickleable(action)
        unpickled = pickle.loads(pickled)

        self.assertTrue(pickleable(action))
        self.assertEqual(unpickled.__dict__, action.__dict__)



from __future__ import annotations
from os import sys
from pathlib import Path
from typing import Optional

from actionpack import Action
from actionpack.action import Name


class Write(Action[Name, int]):
    def __init__(
        self,
        filename: str,
        to_write: str,
        prefix: Optional[str] = None,
        overwrite: bool = False,
        append: bool = False,
        mkdir: bool = False,
    ):
        acceptable_prefix_types = [bytes, str]
        prefix_type, to_write_type = type(prefix), type(to_write)

        if (
            prefix
            and (
                to_write_type not in acceptable_prefix_types
                or prefix_type != to_write_type
            )
        ):
            type_names = ' or '.join([t.__name__ for t in acceptable_prefix_types])
            addendum = f'"{to_write_type.__name__}" given to write with "{prefix_type.__name__}" as prefix.'
            msg = f'Data to write and its prefix must be of types: {type_names}, but {addendum}'
            raise TypeError(msg)

        self.to_write = to_write
        self.prefix = prefix
        self.append = append
        self.overwrite = overwrite
        self.mkdir = mkdir
        self.area = None
        self.path = None

        if filename is Write.STDOUT:
            self.area = sys.stdout
        elif filename is Write.STDERR:
            self.area = sys.stderr
        else:
            self.path = Path(filename)

    def instruction(self) -> str:
        if self.area:
            self.area.write(f"{self.prefix if self.prefix else ''}{self.to_write}")
            self.area.flush()
            return None

        if isinstance(self.to_write, bytes):
            mode, msg = 'ab', self.prefix if self.prefix else b'' + self.to_write
        elif isinstance(self.to_write, str):
            mode, msg = 'a', f"{self.prefix if self.prefix else ''}{self.to_write}"
        else:
            raise TypeError(f'Must be of str or bytes: {self.to_write}')

        if self.mkdir and not self.path.is_dir():
            self.path.resolve().parent.mkdir(parents=True, exist_ok=True)

        if self.append:
            rchar = b'\n' if mode == 'ab' else '\n'
            self.path.open(mode).write(msg + rchar)
        else:
            self.path.write_bytes(msg) if isinstance(msg, bytes) else self.path.write_text(self.to_write)
        return str(self.path.absolute())

    def validate(self) -> Write[Name, int]:
        if self.overwrite and self.append:
            raise ValueError('Cannot overwrite and append simultaneously')
        if self.path and self.path.is_dir():
            raise IsADirectoryError(str(self.path))
        if self.path and self.path.exists() and not self.overwrite and not self.append:
            raise FileExistsError(f'Cannot {str(self)} to {str(self.path)}')
        return self

    class STDOUT:
        pass

    class STDERR:
        pass



from actionpack import Action
from actionpack.action import Name
from actionpack.action import Outcome
from actionpack.utils import Closure


class Call(Action[Name, Outcome]):
    def __init__(self, closure: Closure[Outcome]):
        if not isinstance(closure, Closure):
            raise TypeError('The `closure` parameter must be of type Closure')
        self.closure = closure

    def instruction(self) -> Outcome:
        return self.closure()

    def __repr__(self):
        return f'<Call({self.closure.func.__name__})>'



from typing import TypeVar
from typing import Union

from actionpack import Action
from actionpack.action import Name


T = TypeVar('T')
Outcome = Union[str, T]


class Serialization(Action[Name, Outcome]):
    def __init__(self, schema=None, data=None, inverse=False):
        self.schema = schema
        self.data = data
        self.inverse = inverse

    def validate(self):
        if not self.data:
            raise self.NoDataGiven()
        if not self.schema:
            raise self.NoSchemaGiven()
        return self

    def instruction(self) -> Outcome:
        return self.schema.loads(self.data) if self.inverse else self.schema.dumps(self.data)

    class NoDataGiven(Exception):
        pass

    class NoSchemaGiven(Exception):
        pass



from __future__ import annotations
from string import Template
from time import sleep

from actionpack import Action
from actionpack.action import Result
from actionpack.action import Name
from actionpack.action import Outcome
from actionpack.utils import tally


class RetryPolicy(Action[Name, Outcome]):
    def __init__(
        self,
        action: Action[Name, Outcome],
        max_retries: int,
        delay_between_attempts: int = 0,
        should_record: bool = False
    ):
        self.action = action
        self.max_retries = max_retries
        self.delay_between_attempts = delay_between_attempts
        self.should_record = should_record

        if self.should_record:
            self.attempts: list[Result[Outcome]] = []

    def instruction(self) -> Outcome:
        return self.enact(self.delay_between_attempts)

    def validate(self) -> RetryPolicy:
        try:
            if self.retries >= 0:
                raise RetryPolicy.Expired(f'{str(self.action)} already attempted. Will not perform.')
        except AttributeError:
            pass
        finally:
            return self

    def enact(self, with_delay: int = 0, counter: int = -1) -> Outcome:
        for _tally in tally(1 + self.max_retries):
            attempt = self.action.perform()
            counter += _tally
            self.retries = counter
            if self.should_record:
                self.attempts.append(attempt)
            if attempt.successful:
                return attempt.value
            sleep(with_delay)

        raise RetryPolicy.Expired(f'Max retries exceeded: {self.max_retries}.')

    def __repr__(self):
        tmpl = Template('<$class_name($max_retries x $action_name$delay)>')
        return tmpl.substitute(
            class_name=self.__class__.__name__,
            max_retries=self.max_retries,
            action_name=str(self.action),
            delay='' if not self.delay_between_attempts else f' | {self.delay_between_attempts}s delay'
        )

    class Expired(Exception):
        pass



from actionpack.actions.call import Call
from actionpack.actions.pipeline import Pipeline
from actionpack.actions.make_request import MakeRequest
from actionpack.actions.read import Read
from actionpack.actions.read_input import ReadInput
from actionpack.actions.remove import Remove
from actionpack.actions.retry_policy import RetryPolicy
from actionpack.actions.serialization import Serialization
from actionpack.actions.write import Write


__all__ = [
    'Call',
    'MakeRequest',
    'Pipeline',
    'Read',
    'ReadInput',
    'Remove',
    'RetryPolicy',
    'Serialization',
    'Write',
]



import pickle

from calendar import timegm
from datetime import datetime
from functools import wraps
from typing import Callable
from typing import Dict
from typing import Generic
from typing import Iterable
from typing import List
from typing import Optional
from typing import TypeVar


T = TypeVar('T')


class Closure(Generic[T]):
    def __init__(self, func: Callable[..., T], *args, **kwargs):
        if callable(func) and func.__name__ == '<lambda>':
            raise self.LambdaNotAllowed(repr(func))
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self) -> T:
        return self.func(*self.args, **self.kwargs)

    def __repr__(self):
        return f'<{self.__class__.__name__}({self.func.__name__})|*{self.args}, **{self.kwargs}>'

    def __hash__(self):
        contents = (self.func, self.args, tuple(self.kwargs.items()))
        return hash(contents)

    def __eq__(self, other):
        return hash(self) == hash(other)

    class LambdaNotAllowed(Exception):
        pass


def tally(num=1):
    while (num if num > 0 else -1 * num) > 0:
        yield 1 if num > 0 else -1
        num = num - 1 if num > 0 else num + 1


def pickleable(obj) -> Optional[bytes]:
    try:
        return pickle.dumps(obj)
    except Exception:
        pass


def synchronized(lock):
    def wrap(f):
        @wraps(f)
        def lockedFunction(*args, **kw):
            with lock:
                return f(*args, **kw)
        return lockedFunction
    return wrap


def first(iterable: Iterable):
    return iterable[0]


def last(iterable: Iterable):
    return iterable[-1]


def key_for(value: T, dct: Dict) -> Optional[T]:
    try:
        return list(dct.keys())[list(dct.values()).index(value)]
    except ValueError:
        return None


def swap(lst: List, idx1: int, idx2: int) -> List:
    lst[idx1], lst[idx2] = lst[idx2], lst[idx1]
    return lst


def microsecond_timestamp(clock: Callable[[], datetime] = datetime.utcnow) -> int:
    now = clock()
    return  int(timegm(now.timetuple()) * 1e6 + now.microsecond)

